from omnipy.compute.flow import LinearFlowTemplate
from omnipy.modules.general.tasks import import_directory
from omnipy.modules.json.flows import flatten_nested_json
from omnipy.modules.json.tasks import (convert_dataset_string_to_json,
                                       transpose_dicts_of_lists_of_dicts_2_lists_of_dicts)
from omnipy.modules.pandas.tasks import convert_dataset_list_of_dicts_to_pandas


@LinearFlowTemplate(
    import_directory.refine(
        name='import_json_files_from_dir',
        fixed_params=dict(include_suffixes=('.json',)),
    ),
    convert_dataset_string_to_json,
    transpose_dicts_of_lists_of_dicts_2_lists_of_dicts,
    flatten_nested_json,
    convert_dataset_list_of_dicts_to_pandas,
)
def convert_isa_json_to_relational_tables(dir_path: str):
    ...



import asyncio
from asyncio import Future
from pathlib import Path
import random
import textwrap
import time

import github3
import httpx
from omnipy import runtime
from omnipy.compute.flow import FuncFlowTemplate, LinearFlowTemplate
from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model
from omnipy_examples.chatgpt_words import (literary_styles,
                                           surrealistic_adjectives,
                                           surrealistic_nouns)
from pydantic.main import BaseModel
from reader import Feed, make_reader

biorxiv_feed_url = 'https://connect.biorxiv.org/biorxiv_xml.php?subject=all'
chatgpt_secret_key = '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'

# TODO: Implement omnified way of managing API keys and other secrets

# Models


class PaperInfo(BaseModel):
    id: str = ''
    title: str = ''
    abstract: str = ''


class PaperInfoWithPrompt(PaperInfo):
    prompt: str = ''


class PaperInfoWithPromptAndInterpretation(PaperInfoWithPrompt):
    interpretation: str = ''


class PaperInfoDataset(Dataset[Model[PaperInfo]]):
    ...


class PaperInfoWithPromptDataset(Dataset[Model[PaperInfoWithPrompt]]):
    ...


class PaperInfoWithPromptAndInterpretationDataset(
        Dataset[Model[PaperInfoWithPromptAndInterpretation]]):
    ...


@TaskTemplate()
def get_latest_entries(feed_url: str, max_entries_per_run: int) -> PaperInfoDataset:
    dataset = PaperInfoDataset()

    db_path = Path(runtime.config.root_log.file_log_dir_path).parent.joinpath('db.sqlite')
    reader = make_reader(str(db_path))
    reader.add_feed(feed_url, exist_ok=True)
    reader.update_feeds()

    for i, entry in enumerate(reader.get_entries(read=False)):
        if i >= max_entries_per_run:
            break

        reader.mark_entry_as_read(entry)

        title = entry.title.replace('/', '-')
        dataset[title] = PaperInfo(
            id=entry.id,
            title=title,
            abstract=entry.summary,
        )

    return dataset


def pick_a_word(word_list: list[str]):
    return random.sample(word_list, 1)[0].lower()


@TaskTemplate(iterate_over_data_files=True)
def generate_prompt(
    paper: PaperInfo,
    genres: list[str],
    adjectives: list[str],
    nouns: list[str],
) -> PaperInfoWithPrompt:
    prompt = f'Here follows the abstract of a scientific paper recently published at' \
             f'bioRxiv ({paper.id}):\n\n{paper.abstract}\n\n' \
             f'Based on the above abstract, can you write a text in the literary ' \
             f'style of {pick_a_word(genres)}, where the text has stylistic ' \
             f'qualities best described as {pick_a_word(adjectives)} and {pick_a_word(adjectives)}, ' \
             f'and which also somehow includes a {pick_a_word(nouns)} and a {pick_a_word(nouns)} ' \
             f'multiple times in the discussion?'
    return PaperInfoWithPrompt(
        id=paper.id,
        title=paper.title,
        abstract=paper.abstract,
        prompt=prompt,
    )


@LinearFlowTemplate(
    get_latest_entries.refine(fixed_params=dict(
        feed_url=biorxiv_feed_url,
        max_entries_per_run=5,
    )),
    generate_prompt.refine(
        fixed_params=dict(
            genres=literary_styles,
            adjectives=surrealistic_adjectives,
            nouns=surrealistic_nouns,
        )),
)
async def get_biorxiv_entries_and_generate_prompts() -> PaperInfoWithPromptDataset:
    ...


# get_biorxiv_entries_and_generate_prompts.run()

# Possible simplification?
#
# @TaskTemplate(async_iterate_over_data_files=True, async_run_until_complete=True, async_batch_size=1)
# async def generate_responses(paper: PaperInfo) -> Future:
#     API_KEY = chatgpt_secret_key
#     API_ENDPOINT = 'https://api.openai.com/v1/completions'
#     await async_http_post(endpoint=API_ENDPOINT, headers={'Authorization': f'Bearer {API_KEY}'})


@TaskTemplate()
async def generate_response(paper: PaperInfo) -> Future:
    API_KEY = chatgpt_secret_key
    API_ENDPOINT = 'https://api.openai.com/v1/completions'

    async with httpx.AsyncClient() as client:
        # with httpx.Client() as client:
        response = await client.post(
            # response = client.post(
            API_ENDPOINT,
            headers={
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {API_KEY}',
            },
            json={
                'model': 'text-davinci-003',
                'prompt': paper.prompt,
                'max_tokens': 400,
                'temperature': 0.9,
            },
            timeout=200)
        # return response.json()['choices'][0]['text']
        return response.json()


# Possible simplification of first half of generate_responses_for_all_prompts?
#    responses = generate_responses(papers)


@FuncFlowTemplate
def generate_responses_for_all_prompts(papers: PaperInfoWithPromptDataset):
    loop = asyncio.new_event_loop()
    tasks = [generate_response(paper) for paper in papers.values()]
    # for paper in papers.values():
    #     tasks.append(generate_response(paper))
    #     time.sleep(1)
    responses = loop.run_until_complete(asyncio.wait(tasks))[0] if tasks else []
    response_dict = {key: response.result() for key, response in zip(papers.keys(), responses)}

    dataset = PaperInfoWithPromptAndInterpretationDataset()
    for key, paper in papers.items():
        dataset[key] = PaperInfoWithPromptAndInterpretation(
            id=paper.id,
            title=paper.title,
            abstract=paper.abstract,
            prompt=paper.prompt,
            interpretation=response_dict[key]['choices'][0]['text'],
        )
    return dataset


@TaskTemplate(iterate_over_data_files=True)
def commit_info_to_github_repo(paper: PaperInfoWithPromptAndInterpretation,
                               branch: str = 'main') -> PaperInfoWithPromptAndInterpretation:
    # Authenticate with a personal access token
    gh = github3.login(
        token=
        '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'
    )

    # Get the repository by owner and name
    repo = gh.repository('fairtracks', 'biorxiv_chatgpt')

    filename = paper.id.split('/')[-1].split('?')[0]
    contents = paper.prompt + '\n\nInterpretation of "text-davinci-003":\n' + paper.interpretation
    contents = textwrap.fill(contents, 80, replace_whitespace=False)
    # Create a new file in the repository
    repo.create_file(
        path=f'papers/{filename}.txt',
        message=f'Adding "{paper.title}"',
        content=contents.encode('utf8'),
        branch='wp3_finse_test')

    # Commit the changes with a commit message
    # last_commit = list(repo.commits())[-1]

    # Push the changes to the remote repository
    # last_commit.push()

    return paper


@LinearFlowTemplate(
    get_biorxiv_entries_and_generate_prompts,
    generate_responses_for_all_prompts,
)
def get_chatgpt_interpretation_of_biorxiv_entries() -> PaperInfoWithPromptAndInterpretationDataset:
    ...


@LinearFlowTemplate(
    get_chatgpt_interpretation_of_biorxiv_entries,
    commit_info_to_github_repo,
)
def get_chatgpt_interpretation_of_biorxiv_entries_and_commit(
) -> PaperInfoWithPromptAndInterpretationDataset:
    ...


@FuncFlowTemplate()
def get_chatgpt_interpretation_of_biorxiv_entries_and_commit_loop(
) -> PaperInfoWithPromptAndInterpretationDataset:
    while True:
        papers = get_chatgpt_interpretation_of_biorxiv_entries_and_commit()
        if len(papers) == 0:
            break
        time.sleep(60)


def main():
    # generate_response.run(q)
    # asyncio.run(generate_response.run(q))
    get_chatgpt_interpretation_of_biorxiv_entries_and_commit_loop.run()


# await print(generate_response('What is ChatGPT?'))

if __name__ == '__main__':
    main()
    # import asyncio
    # asyncio.run(main())

# get_latest_entries.run(biorxiv_feed_url)



from omnipy.compute.flow import LinearFlowTemplate
from omnipy.modules.general.tasks import import_directory
from omnipy.modules.json.datasets import JsonDictOfDictsDataset
from omnipy.modules.raw.tasks import modify_datafile_contents

# from omnipy.modules.r_stat import r

# Regex patterns for parsing
#     variable_pattern = re.compile(r"  type discrete \[ \d+ \] \{ (.+) \};\s*")
#     prior_probability_pattern_1 = re.compile(
#         r"probability \( ([^|]+) \) \{\s*")
#     prior_probability_pattern_2 = re.compile(r"  table (.+);\s*")
#     conditional_probability_pattern_1 = (
#         re.compile(r"probability \( (.+) \| (.+) \) \{\s*"))
#     conditional_probability_pattern_2 = re.compile(r"  \((.+)\) (.+);\s*")

# @TaskTemplate
# def import_dag_from_bnlearn(dag_name: str):
#     r('chooseCRANmirror(ind = 1)')
#     r('install.binaries("bnlearn")')
#     r('library(bnlearn)')
#     # r('install.packages("https://www.bnlearn.com/releases/bnlearn_latest.tar.gz", '
#     #   'repos = NULL, type = "source")')


def convert_to_json(contents: str, **kwargs: object):
    contents = contents.replace('\n', '')
    return f'"{contents}"'


@LinearFlowTemplate(
    import_directory.refine(
        name='import_and_convert_bif_files_to_json',
        fixed_params=dict(include_suffixes=('.bif',)),
    ),
    modify_datafile_contents.refine(
        name='modify_bif_files',
        fixed_params=dict(modify_contents_func=convert_to_json),
    ),
)
def import_and_convert_bif_files_to_json(dir_path: str) -> JsonDictOfDictsDataset:
    ...



from omnipy.compute.flow import FuncFlowTemplate, LinearFlowTemplate
from omnipy.compute.task import TaskTemplate
from omnipy.modules.general.tasks import cast_dataset
from omnipy.modules.json.datasets import JsonDataset
from omnipy.modules.json.flows import flatten_nested_json
from omnipy.modules.json.models import JsonDictModel
from omnipy.modules.json.tasks import transpose_dicts_of_lists_of_dicts_2_lists_of_dicts
from omnipy.modules.pandas.models import PandasDataset
from omnipy.modules.pandas.tasks import convert_dataset_list_of_dicts_to_pandas
import pandas as pd
import requests


@TaskTemplate
def import_uniprot():
    HEADERS = {'accept': 'application/json'}
    api_url = 'https://rest.uniprot.org/uniprotkb/search?query=human%20cdc7'
    response = requests.get(api_url, headers=HEADERS)
    if response.status_code == 200:
        dataset = JsonDataset()
        dataset['uniprotkb'] = response.json()
        return dataset
    else:
        raise RuntimeError('No result found')


# @FuncFlowTemplate()
# def import_and_flatten_uniprot() -> Dataset[JsonTableOfStrings]:
#     uniprot_1_ds = import_uniprot()
#     uniprot_2_ds = cast_dataset(uniprot_1_ds, cast_model=JsonDictOfAnyModel)
#     uniprot_3_ds = transpose_dicts_of_lists_of_dicts_2_lists_of_dicts(uniprot_2_ds)
#     uniprot_4_ds = flatten_nested_json_to_list_of_dicts(uniprot_3_ds)
#
#     uniprot_5_ds = cast_dataset(uniprot_4_ds, cast_model=JsonTableOfStrings)
#     return to_pandas(uniprot_5_ds)
#     # return

cast_json = cast_dataset.refine(fixed_params=dict(cast_model=JsonDictModel)),


@LinearFlowTemplate(
    # TODO: Bug running cast_json in linear flow
    #       (probably due to two argument parameters in cast_dataset)
    # TODO: When automatic transformation of task/flow inputs/outputs are implemented,
    #       remove to_pandas calls, here and otherwise
    import_uniprot,  # cast_json,
    transpose_dicts_of_lists_of_dicts_2_lists_of_dicts,
    flatten_nested_json,
    convert_dataset_list_of_dicts_to_pandas)
def import_and_flatten_uniprot() -> PandasDataset:
    ...


@TaskTemplate
def pandas_magic(pandas: PandasDataset) -> PandasDataset:
    #  Get synonym table and clean foreign key
    df_synonym = pandas['results.genes.synonyms']
    df_synonym['_omnipy_ref'] = df_synonym['_omnipy_ref'].str.strip('results.genes.')

    # Get gene table and join with synonym table to get gene foreign id
    df_gene = pandas['results.genes']
    df_merge_1 = pd.merge(
        df_synonym, df_gene, left_on='_omnipy_ref', right_on='_omnipy_id', how='right')
    df_merge_1 = df_merge_1.loc[:, ['value', '_omnipy_ref_y']]
    df_merge_1.columns = ['synomym', '_omnipy_ref']
    df_merge_1['_omnipy_ref'].replace('results.', '', inplace=True, regex=True)

    # Get keywords table and clean foreign key
    df_keywords = pandas['results.keywords']
    df_keywords['_omnipy_ref'].replace('results.', '', inplace=True, regex=True)
    df_keywords = df_keywords.loc[:, ['_omnipy_ref', 'category', 'name']]

    # Merge keywords with synonym
    df_merge_2 = pd.merge(df_merge_1, df_keywords, on='_omnipy_ref', how='right')

    # Get results table for regene name and primary accession
    df_results = pandas['results']
    df_results = df_results.loc[:, ['_omnipy_id', 'primaryAccession', 'uniProtkbId']]
    df_merge_final = pd.merge(
        df_merge_2, df_results, left_on='_omnipy_ref', right_on='_omnipy_id', how='right')

    out_dataset = PandasDataset()
    out_dataset['my_table'] = df_merge_final

    return out_dataset


@FuncFlowTemplate
def import_and_flatten_uniprot_with_magic() -> PandasDataset:
    uniprot_6_ds = import_and_flatten_uniprot()
    uniprot_7_ds = pandas_magic(uniprot_6_ds)
    return uniprot_7_ds


# @TaskTemplate
# def join_a_with_b(pandas_ds: PandasDataset,
#                   table_a_name: str,
#                   a_ref_column: str,
#                   table_b_name: str,
#                   b_id_column: str,
#                   join_table_name: str) -> PandasDataset:
#     ...
#
#
# join_a_with_b(uniprot_6_ds,
#               'results.genes.synonyms',
#               '_omnipy_ref',
#               'results.genes',
#               '_omnipy_id',
#               'merged_table')

# uniprot_7_ds = pandas_magic.run(uniprot_6_ds)

# # output
# serialize_to_tarpacked_json_files('1_uniprot_per_infile_ds', uniprot_1_ds)
# serialize_to_tarpacked_json_files('2_uniprot_per_infile_dict_ds',
# uniprot_2_ds)
# serialize_to_tarpacked_json_files('3_uniprot_nested_list_ds', uniprot_3_ds)
# serialize_to_tarpacked_json_files('4_uniprot_unnested_list_ds', uniprot_4_ds)
# serialize_to_tarpacked_json_files('5_uniprot_tabular_json', uniprot_5_ds)
# serialize_to_tarpacked_csv_files('6_uniprot_tabular_csv', uniprot_5_ds)
#
# serialize_to_tarpacked_csv_files('7_output_csv', uniprot_7_ds)
# results
#     uniProtkbId
#     primaryAccession
# results.keywords
#     name
# results.gene.synonyms
#     value



from omnipy.compute.flow import LinearFlowTemplate
from omnipy.modules.fairtracks.tasks import import_dataset_from_encode
from omnipy.modules.json.flows import flatten_nested_json
from omnipy.modules.pandas.models import PandasDataset
from omnipy.modules.pandas.tasks import convert_dataset_list_of_dicts_to_pandas
from omnipy.modules.tables.tasks import remove_columns

# cast_to_dict_on_top = cast_dataset.refine(
#     name='cast_to_dict_on_top',
#     fixed_params=dict(cast_model=JsonDictOfAnyModel),
# )

# encode_json_pruned =
# encode_json_dict = cast_to_dict_on_top.run(encode_json)

# runtime.config.engine = 'prefect'
# runtime.config.prefect.use_cached_results = True
#
#
# @FuncFlowTemplate
# def import_encode_data_tmpl():
#     encode_json = import_dataset_from_encode(
#         endpoints=[
#             'experiment',
#             'biosample',
#         ],
#         max_data_item_count=25,
#         serialize_as='csv',
#     )
#     encode_json_pruned = remove_columns(
#         encode_json,
#         column_keys_for_data_items=dict(
#             experiment=['audit'],
#             biosample=['audit'],
#         ),
#         serialize_as='csv',
#     )
#     return encode_json_pruned
#
#
# import_encode_data.run()


@LinearFlowTemplate(
    import_dataset_from_encode.refine(
        fixed_params=dict(
            endpoints=[
                'experiment',
                'biosample',
            ],
            max_data_item_count=25,
        )),
    flatten_nested_json,
    remove_columns.refine(
        fixed_params=dict(
            column_keys_for_data_items=dict(
                experiment=['audit'],
                biosample=['audit'],
            ))),
    convert_dataset_list_of_dicts_to_pandas,
)
def import_and_flatten_encode_data() -> PandasDataset:
    ...



from typing import Callable, Optional, TypeAlias

from omnipy.modules.json.datasets import JsonDataset
from omnipy.modules.json.models import JsonModel
from pydantic import BaseModel

IsaJsonModel: TypeAlias = JsonModel
IsaJsonDataset: TypeAlias = JsonDataset
ExtendedIsaJsonDataset: TypeAlias = JsonDataset

JsonPath: TypeAlias = str


class FilteringRule(BaseModel):
    key: JsonPath
    value: str


class TargetRepository(BaseModel):
    name: str
    api_url: str
    filtering_rules: list[FilteringRule]
    file_submitter: Optional[Callable]


TargetRepository(
    name='ENA',
    api_url='https://oajsfkjas',
    filtering_rules=[
        FilteringRule(
            key='investigation.studies.assays.technologyType.annotationValue',
            value='nucleotide sequencing')
    ],
    file_submitter=None)


class Credential(BaseModel):
    token: str


def broker_isa_json_to_multi_repos(dataset: IsaJsonDataset,
                                   target_repositories: list[TargetRepository],
                                   credentials: Credential) -> ExtendedIsaJsonDataset:
    ...



import importlib
from pathlib import Path

from omnipy import runtime
from omnipy.api.enums import ConfigPersistOutputsOptions, ConfigRestoreOutputsOptions, EngineChoice
import typer

app = typer.Typer()
runtime.config.data.interactive_mode = False


def get_path_to_example_data() -> Path:
    ref = importlib.resources.files('omnipy_example_data')
    path: Path
    with importlib.resources.as_file(ref) as path:
        return path.resolve()


installed_example_data_path = get_path_to_example_data()


@app.command()
def dagsim(input_dir: str = installed_example_data_path.joinpath('bif')) -> object:
    from omnipy_examples.dagsim import import_and_convert_bif_files_to_json
    return import_and_convert_bif_files_to_json.run(input_dir)


@app.command()
def encode() -> object:
    from omnipy_examples.encode import import_and_flatten_encode_data
    return import_and_flatten_encode_data.run()


@app.command()
def gff(input_dir: str = installed_example_data_path.joinpath('gff')) -> object:
    from omnipy_examples.gff import import_gff_as_pandas
    return import_gff_as_pandas.run(input_dir)


@app.command()
def isajson(input_dir: str = installed_example_data_path.joinpath('isa-json')) -> object:
    from omnipy_examples.isajson import convert_isa_json_to_relational_tables
    return convert_isa_json_to_relational_tables.run(input_dir)


@app.command()
def uniprot() -> object:
    from omnipy_examples.uniprot import import_and_flatten_uniprot_with_magic
    return import_and_flatten_uniprot_with_magic.run()


@app.command()
def chatgpt() -> object:
    from omnipy_examples.chatgpt import \
        get_chatgpt_interpretation_of_biorxiv_entries_and_commit_loop
    return get_chatgpt_interpretation_of_biorxiv_entries_and_commit_loop.run()


@app.callback()
def main(output_dir: str = runtime.config.job.output_storage.local.persist_data_dir_path,
         engine: EngineChoice = 'local',
         persist_outputs: ConfigPersistOutputsOptions = 'all',
         restore_outputs: ConfigRestoreOutputsOptions = 'disabled'):

    runtime.config.engine = engine
    runtime.config.job.persist_data_dir_path = output_dir
    runtime.config.job.persist_outputs = persist_outputs
    runtime.config.job.restore_outputs = restore_outputs
    runtime.config.data.interactive_mode = False


if __name__ == '__main__':
    app()



from omnipy import runtime
from omnipy_examples.main import chatgpt, isajson
from prefect import flow as prefect_flow


@prefect_flow
def isajson_prefect():
    runtime.config.engine = 'prefect'
    isajson()


@prefect_flow
def chatgpt_prefect():
    runtime.config.engine = 'prefect'
    chatgpt()


if __name__ == "__main__":
    isajson_prefect.serve(name="isajson-prefect-deployment")



literary_styles = [
    'Fiction',
    'Non-fiction',
    'Poetry',
    'Drama',
    'Prose',
    'Satire',
    'Comedy',
    'Tragedy',
    'Epic',
    'Sonnet',
    'Haiku',
    'Lyric',
    'Narrative',
    'Mystery',
    'Horror',
    'Romance',
    'Historical fiction',
    'Science fiction',
    'Fantasy',
    'Realism',
    'Naturalism',
    'Modernism',
    'Postmodernism',
    'Magical realism',
    'Surrealism',
    'Stream of consciousness',
    'Gothic',
    'Dystopian',
    'Bildungsroman',
    'Picaresque',
    'Memoir',
    'Autobiography',
    'Biography',
    'Travel writing',
    'Essay',
    'Journal',
    'Criticism',
    'Manifesto',
    'Satire',
    'Parody',
    'Fable',
    'Allegory',
    'Myth',
    'Legend',
    'Folklore',
    'Fairy tale',
    'Nursery rhyme',
    'Ballad',
    'Limerick',
    'Epistolary',
]
surrealistic_adjectives = [
    'Jumbo',
    'Whispering',
    'Glowing',
    'Inflatable',
    'Cactus',
    'Crispy',
    'Billowing',
    'Bouncing',
    'Rotating',
    'Swirling',
    'Sticky',
    'Blazing',
    'Fizzy',
    'Jagged',
    'Glimmering',
    'Wobbling',
    'Tangled',
    'Hissing',
    'Luminous',
    'Crunchy',
    'Gaseous',
    'Serpentine',
    'Bumpy',
    'Flimsy',
    'Spinning',
    'Sizzling',
    'Blinking',
    'Fluffy',
    'Crinkly',
    'Rustling',
    'Sparkling',
    'Quivering',
    'Zesty',
    'Dangling',
    'Glazed',
    'Drifting',
    'Illuminated',
    'Dizzy',
    'Wavy',
    'Chirping',
    'Smoldering',
    'Quicksilver',
    'Splintered',
    'Gurgling',
    'Wispy',
    'Stuttering',
    'Thunderous',
    'Gossamer',
    'Spiky',
    'Oscillating'
]
surrealistic_nouns = [
    'Lighthouse',
    'Trampoline',
    'Kaleidoscope',
    'Windmill',
    'Dandelion',
    'Waterfall',
    'Wristwatch',
    'Jellyfish',
    'Chandelier',
    'Zucchini',
    'Tambourine',
    'Thunderbolt',
    'Seashell',
    'Saxophone',
    'Hammock',
    'Mandolin',
    'Traffic light',
    'Helicopter',
    'Feather boa',
    'Teapot',
    'Hula hoop',
    'Pinwheel',
    'Folding fan',
    'Ocarina',
    'Slinky',
    'Salt shaker',
    'Harmonica',
    'Teardrop',
    'Flashlight',
    'Disco ball',
    'Electric fan',
    'Barometer',
    'Microscope',
    'Fire hydrant',
    'Kite',
    'Hairbrush',
    'Megaphone',
    'Umbrella',
    'Paper clip',
    'Stethoscope',
    'Vase',
    'Tumbleweed',
    'Hourglass',
    'Cuckoo clock',
    'Sundial',
    'Trombone',
    'Accordion',
    'Tennis racket',
    'Molotov cocktail',
    'Banjo',
    'Candelabra'
]



import os
from typing import List, Optional, Tuple

from omnipy.compute.flow import DagFlowTemplate, FuncFlowTemplate, LinearFlowTemplate
from omnipy.compute.task import TaskTemplate
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model
from omnipy.modules.general.tasks import import_directory, split_dataset
from omnipy.modules.pandas.models import PandasDataset, PandasModel
from omnipy.modules.pandas.tasks import (concat_dataframes_across_datasets,
                                         convert_dataset_csv_to_pandas,
                                         convert_dataset_pandas_to_csv,
                                         extract_columns_as_files)
from omnipy.modules.raw.tasks import modify_all_lines, modify_datafile_contents, modify_each_line
from omnipy.modules.tables.models import JsonTableOfStrings

# Constants

GFF_COLS = ['seqid', 'source', 'type', 'start', 'end', 'score', 'strand', 'phase', 'attributes']
ATTRIB_COL = GFF_COLS[-1]

# Functions


def slice_lines_func(all_lines: List[str],
                     start: Optional[int] = None,
                     end: Optional[int] = None) -> List[str]:
    return all_lines[start:end]


# Tasks


@TaskTemplate()
def attrib_df_names(dataset: Dataset[Model[object]]) -> List[str]:
    return [name for name in dataset.keys() if name.endswith(ATTRIB_COL)]


@TaskTemplate()
def transform_attr_line_to_json(_line_no: int, line: str) -> str:
    items = [item.strip().split('=') for item in line.strip().split(';') if item]
    json_obj_items = [f'"{key}": "{val}"' for key, val in items]
    return f'{{{", ".join(json_obj_items)}}},' + os.linesep


# Flows


@LinearFlowTemplate(
    import_directory.refine(
        name='import_gff_files',
        fixed_params=dict(include_suffixes=('.gff',), model=Model[str]),
        persist_outputs='disabled',
    ),
    modify_all_lines.refine(
        name='slice_lines',
        fixed_params=dict(modify_all_lines_func=slice_lines_func, start=0, end=1000),
        # param_key_map=dict(end='num_lines'),
    ),
    convert_dataset_csv_to_pandas.refine(
        name='convert_gff_to_pandas',
        fixed_params=dict(delimiter='\t', first_row_as_col_names=False, col_names=GFF_COLS),
    ),
    restore_outputs='disabled')
def import_gff_as_pandas(directory: str) -> PandasDataset:
    ...

    # num_lines: Optional[int] = None)


@DagFlowTemplate(
    extract_columns_as_files.refine(
        fixed_params=dict(col_names=[ATTRIB_COL]),
        result_key='dataset',
    ),
    attrib_df_names.refine(result_key='datafile_names_for_b'),
    split_dataset,
)
def extract_attrib_col_as_separate_dataset(
        dataset: PandasDataset) -> Tuple[PandasDataset, PandasDataset]:
    ...


@LinearFlowTemplate(
    convert_dataset_pandas_to_csv.refine(fixed_params=dict(first_row_as_col_names=False)),
    modify_each_line.refine(
        name='transform_all_lines_to_json',
        fixed_params=dict(modify_line_func=transform_attr_line_to_json),
    ),
    modify_datafile_contents.refine(
        name='transform_datafile_start_and_end_to_json',
        fixed_params=dict(modify_contents_func=lambda x: f'[{x[:-2]}]'),
    ),  # Brackets + strip comma and newline from end
)
def convert_attrib_col_to_table(dataset: PandasDataset) -> PandasDataset:
    ...


# import_gff_as_pandas.run('input/gff', num_lines=1000)

#
# @FuncFlowTemplate
# def import_gff_and_convert_to_pandas() -> PandasDataset:
#     data: PandasDataset = import_gff_as_pandas(1000)
#     return data
#
#
# import_gff_and_convert_to_pandas.run()

# data_9_attrib = Dataset[JsonTableOfStrings]()
# data_9_attrib.from_json(data_8_attrib.to_data())
#
# pd_data_7_attrib = PandasDataset()
# pd_data_7_attrib.from_data(data_9_attrib.to_data())
#
# pd_data_10 = concat_dataframes_across_datasets(
#     [pd_data_5_main, pd_data_7_attrib],
#     vertical=False,
# )
# return pd_data_10

# @FuncFlowTemplate()
# def convert_gff_files(data: Dataset[Model[str]]) -> PandasDataset:
#     data = import_directory('input/gff', suffix='.gff', model=Model[str])
#
#     data_2 = slice_lines(data, start=0, end=1000)
#
#     pd_data_3 = convert_dataset_csv_to_pandas(data_2,
#                                               delimiter='\t',
#                                               first_row_as_col_names=False,
#                                               col_names=GFF_COLS)
#
#     pd_data_4 = extract_columns_as_files(pd_data_3, [ATTRIB_COL])
#
#     pd_data_5_main, pd_data_3_attrib = split_dataset(
#         pd_data_4, attrib_df_names(pd_data_4))
#
#     data_6_attrib = to_csv(pd_data_3_attrib, first_row_as_col_names=False)
#
#     data_7_attrib = transform_all_lines_to_json(data_6_attrib)
#
#     data_8_attrib = transform_datafile_start_and_end_to_json(data_7_attrib)
#
#     data_9_attrib = Dataset[JsonTableOfStrings]()
#     data_9_attrib.from_json(data_8_attrib.to_data())
#
#     pd_data_7_attrib = PandasDataset()
#     pd_data_7_attrib.from_data(data_9_attrib.to_data())
#
#     pd_data_10 = concat_dataframes_across_datasets(
#         [pd_data_5_main, pd_data_7_attrib],
#         vertical=False,
#     )
#     return pd_data_10

# @FuncFlowTemplate
# def import_gff_and_convert_to_pandas() -> PandasDataset:
#     data: Dataset[Model[str]] = import_directory('input/gff',
#                                                  suffix='.gff',
#                                                  model=Model[str])
#     return convert_gff_files(data)
#
#
# import_gff_and_convert_to_pandas.run()
